/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.engine.source;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.core.AlertCode;
import com.funambol.framework.engine.*;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.server.store.NotFoundException;
import com.funambol.framework.tools.Base64;

import com.funambol.exchange.DataAccessException;

import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.items.common.manager.ItemManager;
import com.funambol.exchange.items.task.TaskParseException;
import com.funambol.exchange.items.task.manager.TaskManager;
import com.funambol.exchange.items.task.model.Task;
import com.funambol.exchange.items.calendar.model.Calendar;

import com.funambol.exchange.util.DateTools;
import com.funambol.exchange.util.StringTools;
import com.funambol.exchange.util.Def;
import com.funambol.framework.engine.source.SyncSource;

/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server tasks items
 *
 * @version $Id: ExchangeTaskSyncSource.java,v 1.1 2006/08/08 07:52:15 gibi Exp $
 */
public class ExchangeTaskSyncSource extends ExchangeSyncSource
        implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // ------------------------------------------------------------ Private data

    private TaskManager tm = null;

    // ------------------------------------------------------------ Constructors

    public ExchangeTaskSyncSource() {
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * init manager
     */
    public void initManager()
            throws SyncSourceException {

        try {

            this.tm = new TaskManager(this.host,
                    this.port,
                    this.exchangeHost);

        } catch (Exception e) {

            throw new SyncSourceException("Error begin sync: " +
                    e.getMessage(), e);
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem addSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Task task = null;
        Task newTask = null;

        String itemKey = null;
        String href = null;

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // created
            // default href = this.CurrentTimesMillis()
            //
            href = String.valueOf(System.currentTimeMillis()) +
                    EXCHANGE_HREF_EXTENSION;

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " is new: adding");
            }

            task = getTaskFromSyncItem(syncItem);

            task.setHref(href);

            newTask = this.tm.setTask(task,
                    username,
                    credentials,
                    exchangeFolder);

            String date = null;
            String dueDate = null;

            if (newTask.getDate() != null) {
                date = DateTools.dateToPDIDate(newTask.getDate());
            } else {
                date = "";
            }

            if (newTask.getDueDate() != null) {
                dueDate = DateTools.dateToPDIDate(newTask.getDueDate());
            } else {
                dueDate = "";
            }

            long crc = StringTools.getCRCFromString(newTask.getSubject() +
                    date +
                    dueDate);

            toSyncNewItems(newTask.getId(), crc, newTask.getLastModified());
            addInExchangeItemsKeys(newTask.getId());

            syncItem = getSyncItem(newTask);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Exchange SyncSource " +
                        sourceURI +
                        " - set syncItem" +
                        syncItem.getKey().getKeyAsString());
            }

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "addSyncItem", e);
            }

            if (newTask != null) {
                toSyncErrorNewItems(newTask.getId());
            }

            throw new SyncSourceException("Error setting the item " +
                    syncItem + " " + e.getMessage(), e);
        }
    }

    /**
     * @see SyncSource
     */
    public SyncItem updateSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Task task = null;
        Task updateTask = null;

        String itemKey = null;
        String href = null;

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // create href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey, principalId);

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " already exists: updating");
            }

            task = getTaskFromSyncItem(syncItem);

            task.setHref(href);

            updateTask = this.tm.setTask(task,
                    username,
                    credentials,
                    exchangeFolder);

            String date = null;
            String dueDate = null;

            if (updateTask.getDate() != null) {
                date = DateTools.dateToPDIDate(updateTask.getDate());
            } else {
                date = "";
            }

            if (updateTask.getDueDate() != null) {
                dueDate = DateTools.dateToPDIDate(updateTask.getDueDate());
            } else {
                dueDate = "";
            }

            long crc = StringTools.getCRCFromString(updateTask.getSubject() +
                    date +
                    dueDate);

            toSyncNewItems(updateTask.getId(), crc, updateTask.getLastModified());

            syncItem = getSyncItem(updateTask);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Exchange SyncSource " +
                        sourceURI +
                        " - set syncItem" +
                        syncItem.getKey().getKeyAsString());
            }

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "updateSyncItem", e);
            }

            if (updateTask != null) {
                toSyncErrorUpdateItems(updateTask.getId());
            }

            throw new SyncSourceException("Error setting the item " +
                    syncItem + " " + e.getMessage(), e);
        }
    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromId(SyncItemKey syncItemKey)
            throws SyncSourceException {

        Task task = null;
        String id = null;
        SyncItem syncItem = null;

        id = syncItemKey.getKeyAsString();

        //
        // search syncItem in cache
        //
        syncItem = getSyncItemById(id);

        if (syncItem != null) {
            return syncItem;
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe
                    ("Note with id " + id + " not found while reading Exchange cache");
        }

        try {

            task = this.tm.getTaskById(username,
                    credentials,
                    id,
                    this.exchangeFolder);
        } catch (Exception e) {

            Throwable previous = e.getCause();

            if (previous instanceof NotFoundException) {

                Logger log = Sync4jLogger.getLogger(Def.LOGGER_NAME);
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Note not found while reading Exchange database: " +
                                    e.getMessage());
                }

                Sync4jLogger.
                        getLogger(Def.LOGGER_NAME).
                        throwing(getClass().
                                getName(), "readExchangeDatabase", e);

            } else {

                if (log.isLoggable(Level.FINEST)) {
                    log.severe("SyncSource error: " + e.getMessage());
                    log.throwing(getClass().getName(), "getSyncItemFromId", e);
                }

                throw new SyncSourceException("Error reading items", e);
            }
        }

        if (task != null) {
            task.setState(getSyncItemStatus(task.getId()));
            syncItem = getSyncItem(task);
        }

        return syncItem;
    }


    /**
     * @see SyncSource
     */
    public void removeSyncItem(SyncItemKey itemKey, Timestamp time, boolean softDelete)
            throws SyncSourceException {

        if (!softDelete) {
            removeSyncItem(itemKey);
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItem)
            throws SyncSourceException {

        Item[] items = null;

        try {
            Task task = getTaskFromSyncItem(syncItem);
            items = this.tm.getTwinItems(task,
                    principalId,
                    this.sourceURI);
        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getSyncItemKeysFromTwin", e);
            }

            throw new SyncSourceException("Error getting twin", e);
        }

        return getSyncItemKeysFromItems(items);
    }


    /**
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return items array
     */
    public Item[] getExchangeItems(String username,
                                   String credentials,
                                   String exchangeFolder,
                                   String[] ids,
                                   boolean isBeginSync)
            throws SyncSourceException {

        Item  [] items = null;
        Task  [] tasks = null;

        try {

            //
            // is not slow sync or called from endSync
            // load calendar with only fields for twin setting
            //
            if (((syncMode != AlertCode.SLOW) &&
                    (syncMode != AlertCode.REFRESH_FROM_SERVER)) ||
                    !isBeginSync) {

                tasks = this.tm.getTasksForTwin(username,
                        credentials,
                        ids,
                        exchangeFolder);


            } else {

                tasks = this.tm.getTasks(username,
                        credentials,
                        ids,
                        exchangeFolder);

                addToCacheSyncItems(getSyncItems(tasks));

            }

            items = getItems(tasks);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return items;

    }

    /**
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     * @return SyncItem array
     */
    public SyncItem[] getSyncItems(String username,
                                   String credentials,
                                   String[] ids,
                                   String exchangeFolder)
            throws SyncSourceException {

        SyncItem [] syncItems = null;
        Task     [] tasks = null;

        try {

            tasks = this.tm.getTasks(username,
                    credentials,
                    ids,
                    exchangeFolder);

            syncItems = getSyncItems(tasks);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return syncItems;

    }

    /**
     * Delete all tasks from the Exchange server
     *
     * @see ExchangeSyncSource
     */
    public void deleteAll()
            throws SyncSourceException {
        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Removing all tasks from Exchange for user " + this.username + ".");
        }

        Task[] allExcTasks;
        try {
            allExcTasks = tm.getTasks(this.username, this.credentials, new String[0], this.getExchangeFolder());
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error removing all tasks from Exchange.");
            }
            throw new SyncSourceException(e);
        }

        for (int i = 0; i < allExcTasks.length; i++) {
            try {
                tm.removeTask(allExcTasks[i], this.username, this.credentials, this.exchangeFolder);
            }
            catch (DataAccessException e) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.log(Level.SEVERE, "Error removing task with id " +
                            allExcTasks[i].getId() + " from Exchange.");
                }
                if (log.isLoggable(Level.INFO)) {
                    log.log(Level.INFO, e.getMessage(), e);
                }

            }
        }

        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Clearing cached tasks for source uri " +
                    this.sourceURI + " and principal " + this.principalId);
        }
        try {
            new ItemManager().deleteItems(tm.getAllItems(this.sourceURI, this.principalId),
                    this.sourceURI, this.principalId);
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error while clearing cached tasks: " + e.getMessage() + ".");
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, e.getMessage(), e);
            }
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Create SyncItem array from Task array
     *
     * @param tasks
     * @return array of SyncItem
     * @throws SyncSourceException
     */
    private SyncItem[] getSyncItems(Task[] tasks)
            throws SyncSourceException {

        SyncItem[] syncItems = null;

        syncItems = new SyncItem[tasks.length];

        for (int i = 0, l = syncItems.length; i < l; i++) {
            syncItems[i] = getSyncItem(tasks[i]);
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Task
     *
     * @param task
     * @return SyncItem
     * @throws SyncSourceException
     */
    private SyncItem getSyncItem(Task task)
            throws SyncSourceException {

        SyncItem syncItem = null;

        syncItem = new SyncItemImpl(this,
                task.getId(),
                task.getState());

        try {
            syncItem.setContent(task.toXml().getBytes());
            syncItem.setType(this.getType());

        } catch (TaskParseException e) {
            throw new SyncSourceException("Error setting the item "
                    + syncItem
                    , e
            );
        }

        return syncItem;

    }

    /**
     * Create SyncItem array from Task array
     *
     * @param tasks
     * @param state
     * @return SyncItem array
     * @throws SyncSourceException
     */
    private SyncItem[] getSyncItems(Task[] tasks,
                                    char state)
            throws SyncSourceException {

        SyncItem[] syncItems = null;

        int l = tasks.length;

        syncItems = new SyncItem [l];

        for (int i = 0; i < l; i++) {
            tasks[i].setState(state);
            syncItems[i] = getSyncItem(tasks[i]);
        }

        return syncItems;

    }

    /**
     * Create Item array from Task array
     *
     * @param tasks
     * @return array of SyncItem
     * @throws SyncSourceException
     */
    private Item[] getItems(Task[] tasks)
            throws SyncSourceException {

        Item[] items = null;

        items = new Item[tasks.length];

        for (int i = 0, l = items.length; i < l; i++) {
            items[i] = getItem(tasks[i]);
        }

        return items;

    }

    /**
     * Create Item from Task
     *
     * @param task
     * @return item
     * @throws SyncSourceException
     */
    private Item getItem(Task task)
            throws SyncSourceException {

        String date = null;
        String dueDate = null;

        Item item = null;

        try {

            if (task.getDate() != null) {
                date = DateTools.dateToPDIDate(task.getDate());
            } else {
                date = "";
            }

            if (task.getDueDate() != null) {
                dueDate = DateTools.dateToPDIDate(task.getDueDate());
            } else {
                dueDate = "";
            }

        } catch (Exception e) {
            throw new SyncSourceException(e.getMessage());
        }

        long crc = StringTools.getCRCFromString(task.getSubject() +
                date +
                dueDate);


        item = new Item(task.getId(),
                task.getHref(),
                task.getLastModified(),
                crc);

        return item;

    }

    /**
     * Creates a new Task from the given syncItem
     *
     * @param syncItem SyncItem
     * @return Task
     */
    private Task getTaskFromSyncItem(SyncItem syncItem)
            throws Exception {

        String content = null;
        Task task = null;

        String itemKey = syncItem.getKey().getKeyAsString();

        byte[] itemContent = (byte[]) syncItem.getContent();

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        content = new String(itemContent);

        Date t = syncItem.getTimestamp();

        task = new Task(itemKey, content, t);

        return task;
    }

    /**
     *
     */
    private void removeSyncItem(SyncItemKey itemKey)
            throws SyncSourceException {

        Task task = null;

        String id = null;
        String href = null;

        id = itemKey.getKeyAsString();

        href = getHref(id, principalId);

        try {
            if (href != null) {
                task = new Task(id);
                task.setHref(href);

                this.tm.removeTask(task,
                        username,
                        credentials,
                        exchangeFolder);
                toSyncRemoveItems(task.getId());
                removeInExchangeItemsKeys(task.getId());
            } else {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Remove task failed. Task with " + id + " not found while reading fnbl_exch_cache");
                }

                toSyncErrorRemoveItems(id);
            }
        } catch (Exception e) {

            toSyncErrorRemoveItems(itemKey.getKeyAsString());

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "removeSyncItem", e);
            }

            throw new SyncSourceException("Error reading items", e);
        }

    }

}
