/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.httptransport;

import java.io.*;

import java.net.Socket;

import java.util.logging.Level;
import java.util.logging.Logger;
import java.security.KeyStore;
import java.security.KeyStoreException;

import com.funambol.framework.logging.Sync4jLogger;

import com.funambol.exchange.util.StringTools;
import com.funambol.exchange.util.Def;
import com.funambol.exchange.admin.ExchangeConnectorConfig;
import com.funambol.exchange.exception.ExchangeConfigException;

import javax.net.ssl.*;

/**
 * This class provide methods to access exchange server datastore
 * by socket connection
 *
 * @version $Id: WebDavHttpTransport.java,v 1.1 2006/08/08 07:51:36 gibi Exp $
 **/
public class WebDavHttpTransport {

    //---------------------------------------------------------------- Constants

    private static final int SIZE_INPUT_BUFFER = 4096;
    private static final String NEW_LINE = "\r\n" ;

    //------------------------------------------------------------- Private data

    private String  host          = null ;
    private int     port          = 0    ;
    private String  exchangeHost  = null ;
    private boolean useSSL        = false;
    private String keyStoreFileName = null;
    private  String keyStorePassword = null;
    protected Logger log = null  ;

    public static int numberCalls = 0;

    //-------------------------------------------------------------- Costructors

    public WebDavHttpTransport(String host         ,
                               int    port         ,
                               String exchangeHost)
    {

        log = Sync4jLogger.getLogger(Def.LOGGER_NAME);

        this.host         = host         ;
        this.port         = port         ;
        this.exchangeHost = exchangeHost ;
        try{
            this.useSSL = ExchangeConnectorConfig.getConfigInstance().getUseSSL();
            this.keyStoreFileName = ExchangeConnectorConfig.getConfigInstance().getKeyStoreFileName();
            this.keyStorePassword = ExchangeConnectorConfig.getConfigInstance().getKeyStorePassword();
        }
        catch(ExchangeConfigException e){log.log(Level.SEVERE, e.toString(), e);}
    }

    //----------------------------------------------------------- Public methods

    /**
     * Send request to exchange server
     * and get response.
     *
     * @param webDavCommand the first row of the header
     * @param credentials the credential
     * @param request the request
     * @param encoding the encoding to use to encode the request and the response
     * @return response
     *
     * @throws IOException
     **/
    public String sendRequest (String webDavCommand,
                               String credentials,
                               String request,
                               String encoding)

        throws IOException, java.security.KeyStoreException
    {

        Socket        s            = null ;
        OutputStream  out          = null ;
        InputStream   in           = null ;

        String        fromExchange;
        String        toExchange;


        StringBuffer temp = new StringBuffer(request);
        temp.append(NEW_LINE).append(NEW_LINE);
        request = temp.toString();

        StringBuffer requestToSend = new StringBuffer(webDavCommand);
        requestToSend.append("Host: ").append(exchangeHost).append(NEW_LINE);
        requestToSend.append("Content-Type: text/xml; charset=\"").append(encoding).append("\"").append(NEW_LINE);
        requestToSend.append("Content-Length: ").append(request.getBytes(encoding).length).append(NEW_LINE);
        requestToSend.append("Authorization: Basic ").append(credentials).append(NEW_LINE);
        requestToSend.append(NEW_LINE);

        requestToSend.append(request);

        toExchange = requestToSend.toString();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Message to Exchange Server:" + toExchange);
        }

        try {

            byte[] bRequest = toExchange.getBytes(encoding);

            if(useSSL)
            {
                if(log.isLoggable(Level.INFO))
                {
                    log.info("Using SSL, keystore file: "+keyStoreFileName+", password: "+ keyStorePassword);
                }
                KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
                InputStream keyStoreIS = new BufferedInputStream(new FileInputStream(keyStoreFileName));
                try
                {
                    keyStore.load(keyStoreIS, keyStorePassword.toCharArray());
                }
                catch(java.security.NoSuchAlgorithmException e)
                {
                    throw new KeyStoreException(e.getMessage());
                }
                catch(java.security.cert.CertificateException e)
                {
                    throw  new KeyStoreException(e.getMessage());
                }
                // Get the TrustManagers from the keystore
                TrustManager[] trustedCerts;
                TrustManagerFactory trustManagerFactory;
                try
                {
                    trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
                }
                catch(java.security.NoSuchAlgorithmException e)
                {
                    throw new KeyStoreException(e.getMessage());
                }
                trustManagerFactory.init(keyStore);
                trustedCerts = trustManagerFactory.getTrustManagers();
                // Install the all-trusting trust manager

                SSLContext sslContext;
                try
                {
                    sslContext = SSLContext.getInstance("SSL");
                    sslContext.init(null, trustedCerts, new java.security.SecureRandom());
                }
                catch(java.security.NoSuchAlgorithmException e)
                {
                    throw new IOException(e.getMessage());
                }
                catch(java.security.KeyManagementException e)
                {
                    throw  new IOException(e.getMessage());
                }


                HttpsURLConnection.setDefaultSSLSocketFactory(sslContext.getSocketFactory());

                SSLSocketFactory  ssFactory = HttpsURLConnection.getDefaultSSLSocketFactory();
                SSLSocket secureSocket = (SSLSocket) ssFactory.createSocket(this.host, this.port);
                secureSocket.startHandshake();
                s = secureSocket;

            }
            else
            {
                if(log.isLoggable(Level.INFO))
                {
                    log.info("not using ssl");
                }
                s = new Socket(this.host,
                               this.port);
            }

            long timeInitConnection = System.currentTimeMillis();

            in  = s.getInputStream ()  ;
            out = s.getOutputStream() ;

            out.write(bRequest);
            out.flush();

            byte[] bResponse = readContent(in);
            in.close();

            long endConnectionTime = System.currentTimeMillis();
            long elapsedTime        = endConnectionTime - timeInitConnection;

            fromExchange = new String(bResponse, encoding);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("WebDav response time - start: " +
                           timeInitConnection               +
                           " - end: "                       +
                           endConnectionTime                +
                           " - elapsed: "                   +
                           elapsedTime                       );

                log.finest("Message from Exchange Server: " + fromExchange);

            }

            return StringTools.unescapeXml(fromExchange);

        } finally {

            if (out != null) {
                out.close () ;
            }
            if (in != null)  {
                in.close  () ;
            }
            if (s != null)   {
                s.close   () ;
            }
        }

    }

    /**
     * Reads the content from the inputstream.
     *
     * @param in the request InputStream
     *
     * @throws IOException in case of errors
     */
    protected byte[] readContent(final InputStream in) throws IOException {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        byte[] buf = new byte[SIZE_INPUT_BUFFER];

        int c = 0;
        int b;
        while ((c < buf.length) && (b = in.read(buf, c, buf.length-c)) >= 0) {
            c+=b;
            if (c == SIZE_INPUT_BUFFER) {
                bout.write(buf);
                buf = new byte[SIZE_INPUT_BUFFER];
                c = 0;
            }
        }
        if (c != 0) {
            bout.write(buf, 0, c);
        }
        return bout.toByteArray();
    }

}