/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.items.note.dao;

import java.text.MessageFormat;

import java.util.ArrayList;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.httptransport.WebDavHttpTransport;
import com.funambol.exchange.items.note.model.Note;
import com.funambol.exchange.items.common.dao.ItemDAO;
import com.funambol.exchange.xml.XmlParser;
import com.funambol.exchange.xml.XmlParseException;
import com.funambol.exchange.util.DateTools;
import com.funambol.exchange.util.StringTools;

/*
 * This class implements methods to access note data
 * in exchange server datastore
 *
 * @version $Id: NoteDAO.java,v 1.1 2006/08/08 07:49:54 gibi Exp $
 *
 **/

public class NoteDAO extends ItemDAO {

    //---------------------------------------------------------------- Constants

    private static final String WEBDAV_MSG_PROPPATCH_UPDATE_NOTE =
            "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n" +
                    "<D:propertyupdate xmlns:D=\"DAV:\" " +
                    "xmlns:EX=\"http://schemas.microsoft.com/exchange/\" " +
                    "xmlns:HN=\"urn:schemas:httpmail:\">\n" +
                    "<D:set>\n" +
                    " <D:prop>\n" +
                    "  <D:contentclass>urn:content-classes:person</D:contentclass>\n" +
                    "  <EX:outlookmessageclass>IPM.StickyNote</EX:outlookmessageclass>\n" +
                    "{0}" +
                    " </D:prop>\n" +
                    "</D:set>\n" +
                    "</D:propertyupdate>";

    private static final String WEBDAV_MSG_SELECT_NOTES =
            "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n" +
                    "<D:searchrequest xmlns:D =\"DAV:\">\n" +
                    "<D:sql>\n" +
                    "Select " +
                    "\"urn:schemas:httpmail:textdescription\" AS textdescription, " +
                    "\"urn:schemas:httpmail:subject\" AS subject, " +
                    "\"urn:schemas:httpmail:date\" AS date, " +
                    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, " +
                    "\"DAV:getlastmodified\" AS getlastmodified, " +
                    "\"DAV:href\" AS href, " +
                    "\"DAV:isfolder\" AS isfolder " +
                    "FROM \"/{0}/{1}/{2}\"  " +
                    "{3} " +
                    "</D:sql>\n" +
                    "</D:searchrequest>";

    private static final String WEBDAV_MSG_SELECT_NOTES_FOR_TWIN =
            "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n" +
                    "<D:searchrequest xmlns:D =\"DAV:\">\n" +
                    "<D:sql>\n" +
                    "Select " +
                    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, " +
                    "\"DAV:getlastmodified\" AS getlastmodified, " +
                    "\"DAV:creationdate\" AS creationdate, " +
                    "\"DAV:href\" AS href, " +
                    "\"urn:schemas:httpmail:subject\" AS subject, " +
                    "\"urn:schemas:httpmail:textdescription\" AS textdescription, " +
                    "\"DAV:isfolder\" AS isfolder " +
                    "FROM \"/{0}/{1}/{2}\"  " +
                    "{3}" +
                    "</D:sql>\n" +
                    "</D:searchrequest>";

    public static final String TAG_DATE = "date";
    public static final String TAG_DESCRIPTION =
            "textdescription";
    public static final String TAG_RESPONSE = "a:response";
    public static final String TAG_SUBJECT = "subject";
    public static final String TAG_REPLUID = "repluid";
    public static final String TAG_LAST_MODIFIED = "getlastmodified";

    //------------------------------------------------------------- Private data

    private WebDavHttpTransport webDavHttp = null;

    //-------------------------------------------------------------- Costructors

    public NoteDAO(String host,
                   int port,
                   String exchangeServer)
            throws DataAccessException {

        this.webDavHttp = new WebDavHttpTransport(host,
                port,
                exchangeServer);

    }
    //----------------------------------------------------------- Public methods

    /**
     * insert / update note
     *
     * @param note
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return new / updated note
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Note setNote(Note note,
                        String username,
                        String credentials,
                        String exchangeFolder,
                        boolean ignoreSubject)
            throws DataAccessException {

        String response;

        String id;

        String webDavNoteMsg;
        String webDavHeaderMsg;

        String server;
        String resource;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        if (ignoreSubject) {
            Note[] serverNotes;
            serverNotes = getNotes(username, credentials, new String[]{note.getId()},exchangeFolder);
            if (serverNotes!=null && serverNotes.length>0) {
                note.setSubject(serverNotes[0].getSubject());
            }
        }

        webDavNoteMsg = getUpdateNoteWebDavRequest(note);

        webDavHeaderMsg = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPPATCH,
                new Object[]{"/" + server + "/" + username + "/" + resource, note.getHref()}
        );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavNoteMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new
                    DataAccessException("Error getting Exchange server response", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " getting Exchange note",
                    e);

        }

        try {
            id = XmlParser.getRuidFromResponse(response);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error getting note id", e);
        }

        note.setId(id);

        return note;
    }

    /**
     * Delete a <i>Note</i> from Exchange server.
     *
     * @param note
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public void removeNote(Note note,
                           String username,
                           String credentials,
                           String exchangeFolder)
            throws DataAccessException {

        String webDavNoteMsg;
        String webDavHeaderMsg;
        String response;

        String server;
        String resource;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        webDavNoteMsg = "";

        webDavHeaderMsg =
                MessageFormat.format(ItemDAO.WEBDAV_HEADER_REMOVE,
                        new Object[]{
                                "/" +
                                        server +
                                        "/" +
                                        username +
                                        "/" +
                                        resource,
                                note.getHref()
                        }
                );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavNoteMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new
                    DataAccessException("Error getting Exchange server response", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " removing exchange note",
                    e);

        }

    }

    /**
     * get notes from Exchange Server
     *
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     * @return array of find notes
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Note[] getNotes(String username,
                           String credentials,
                           String[]  ids,
                           String exchangeFolder)
            throws DataAccessException {

        Note[] exchangeNotes;

        String clause;

        clause = getClause(ids);

        exchangeNotes = getNotes(username, credentials, clause, exchangeFolder);

        return exchangeNotes;
    }

    /**
     * get Note array with setting fields only for twins
     *
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return Note array with setting fields only for twins
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Note[] getNotesForTwin(String username,
                                  String credentials,
                                  String[] ids,
                                  String exchangeFolder)
            throws DataAccessException {

        ArrayList notes;
        Note n;

        String[]  props;
        String prop;
        String[]  resp;

        String replUid;
        String uri;
        String lastUpdate;
        String textDescription;
        String subject;
        String isFolder;
        String response;
        String webDavBodyMsg;
        String webDavHeaderMsg;
        String server;
        String resource;

        int count = 0;

        String clause;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        clause = getClause(ids);

        webDavBodyMsg =
                MessageFormat.format(WEBDAV_MSG_SELECT_NOTES_FOR_TWIN,
                        new Object[]{server, username, resource, clause});

        webDavHeaderMsg =
                MessageFormat.format(WEBDAV_HEADER_SELECT,

                        new Object[]{
                                "/" +
                                        server +
                                        "/" +
                                        username +
                                        "/" +
                                        resource
                        }
                );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg,
                    credentials,
                    webDavBodyMsg,
                    FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error getting Exchange server response",
                    e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " reading exchange items",
                    e);
        }

        resp = new String[]{response};

        try {
            props = XmlParser.getXMLTag(resp, TAG_RESPONSE);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing item", e);
        }

        notes = new ArrayList();

        for (int i = 0, l = props.length; i < l; i++) {

            try {

                prop = props[i];
                isFolder = XmlParser.getXMLInitTagValue(prop, TAG_IS_FOLDER);
                replUid = XmlParser.getXMLTagValue(prop, TAG_REPLUID);
                uri = XmlParser.getXMLTagValue(prop, TAG_HREF);
                lastUpdate = XmlParser.getXMLInitTagValue(prop, TAG_LAST_MODIFIED);
                subject = XmlParser.getXMLInitTagValue(prop, "subject");
                textDescription = XmlParser.getXMLInitTagValue(prop, "textdescription");

                n = new Note(getIdFromReplUid(replUid), DateTools.webDavTagToDate(lastUpdate));

                n.setHref(XmlParser.getHrefFromUri(uri));

                n.setSubject(subject);

                n.setTextDescription(textDescription);

                if (PROP_NO_FOLDER.equals(isFolder)) {
                    notes.add(n);
                    count ++;
                }

            } catch (Exception e) {
                throw new DataAccessException("Error parsing item", e);
            }

        }

        return (Note[]) notes.toArray(new Note[count]);
    }


    /**
     * get String array of notes keys from Exchange Server
     *
     * @param username
     * @param credentials
     * @param fields
     * @param values
     * @param exchangeFolder
     * @return String array of notes keys from Exchange Server
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    public Note[] getNotes(String username,
                           String credentials,
                           String fields[],
                           Object values[],
                           String exchangeFolder) throws DataAccessException {

        String clause;

        clause = getClause(fields, values);

        return getNotes(username, credentials, clause, exchangeFolder);
    }

    //---------------------------------------------------------- Private methods

    /**
     * Create a <i>Note</i> array object from WebDav msg.
     *
     * @param webDavMsg
     * @param webDavMsg
     * @return a newly create <i>Note</i> object initialized with the
     *         fields in the result set
     */
    private Note[] getNotesFromWebDavMsg(String webDavMsg)
            throws DataAccessException {

        ArrayList notes;
        String []  resps;
        String []  msg;

        String response;
        String id;
        String replUid;

        String isFolder;
        int count = 0;

        msg = new String[]{webDavMsg};

        notes = new ArrayList();

        try {

            resps = XmlParser.getXMLTag(msg, TAG_RESPONSE);

            for (int i = 0, l = resps.length; i < l; i++) {

                response = resps[i];
                replUid = XmlParser.getXMLInitTagValue(response, TAG_REPLUID);
                id = getIdFromReplUid(replUid);

                isFolder = XmlParser.getXMLInitTagValue(response,
                        TAG_IS_FOLDER);
                if (PROP_NO_FOLDER.equals(isFolder)) {
                    notes.add(getNoteFromResponseTag(id, response));
                    count ++;
                }
            }

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing note item", e);
        }

        return (Note[]) notes.toArray(new Note[count]);

    }

    /**
     * Create a <i>Note</i> object from response tag.
     *
     * @param id
     * @param msgResponse
     * @return a newly create <i>Note</i> object initialized with the
     *         fields in the result set
     */
    private Note getNoteFromResponseTag(String id, String msgResponse)
            throws DataAccessException {

        Note n;
        String lastUpdate;
        String date;
        String subject;
        String uri;

        n = new Note(id);

        try {

            uri = XmlParser.getXMLTagValue
                    (msgResponse, TAG_HREF);

            n.setHref(XmlParser.getHrefFromUri(uri));

            n.setTextDescription(StringTools.escapeXml(XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_DESCRIPTION)));

            subject = XmlParser.getXMLInitTagValue(msgResponse, TAG_SUBJECT);

            n.setSubject(subject);

            date = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_DATE);

            lastUpdate = XmlParser.getXMLInitTagValue
                    (msgResponse, TAG_LAST_MODIFIED);

            n.setDate(DateTools.webDavTagToDate(date));

            n.setLastModified(DateTools.webDavTagToDate
                    (lastUpdate));

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing note item", e);
        }

        return n;
    }

    /**
     * get tasks from Exchange Server with the given clause
     *
     * @param username
     * @param credentials
     * @param clause
     * @param exchangeFolder
     * @return array of find task
     * @throws com.funambol.exchange.DataAccessException
     *
     */
    private Note[] getNotes(String username,
                            String credentials,
                            String clause,
                            String exchangeFolder) throws DataAccessException {

        Note[] exchangeNotes;

        String response;
        String webDavNoteMsg;
        String webDavHeaderMsg;

        String server;
        String resource;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        webDavNoteMsg =
                MessageFormat.format(WEBDAV_MSG_SELECT_NOTES,
                        new Object[]{server, username, resource, clause});


        webDavHeaderMsg =
                MessageFormat.format(ItemDAO.WEBDAV_HEADER_SELECT,

                        new Object[]{
                                "/" +
                                        server +
                                        "/" +
                                        username +
                                        "/" +
                                        resource
                        }
                );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavNoteMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new
                    DataAccessException("Error getting Exchange server response", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                    username +
                    " URI " +
                    exchangeFolder +
                    " getting exchange note",
                    e);

        }

        exchangeNotes = getNotesFromWebDavMsg(response);
        return exchangeNotes;
    }

    /**
     * Build a clause
     * to webdav select request
     * between ids item array
     *
     * @param fields
     * @param values
     * @return webdav select clause
     */
    private static String getClause(String[] fields, Object[] values) {

        StringBuffer clause = new StringBuffer();

        int l = fields.length;

        if (l == 0) {
            return "";
        }

        clause.append("where ");

        StringBuffer tmp;
        String valueTmp;

        for (int i = 0; i < l; i++) {
            tmp = new StringBuffer();

            if (TAG_SUBJECT.equals(fields[i])) {
                tmp.append("\"urn:schemas:httpmail:subject\"");

                if (values[i] != null) {
                    tmp.append("='");
                    valueTmp = String.valueOf(values[i]);
                    valueTmp = StringTools.escapeXml(valueTmp.replaceAll("'", "''"));
                    tmp.append(valueTmp).append("'");
                } else {
                    tmp.append(" is null ");
                }

            }

            if (i != 0 && tmp.length() != 0) {
                clause.append(" and ").append(tmp);
            } else if (tmp.length() != 0) {
                clause.append(tmp);
            }

        }

        clause.append("\r\n");

        return clause.toString();
    }


    /**
     * <p>Get the update webDav request for a note.
     *
     * @param note
     * @return the update webDav request
     */
    private static String getUpdateNoteWebDavRequest(Note note) {
        StringBuffer fields;
        String subject, body, date;
        fields = new StringBuffer();

        subject = note.getSubject();
        body = note.getTextDescription();
        date = DateTools.dateToWebDavTag(note.getDate());

        if (subject != null) {
            fields.append(" <HN:subject>").append(StringTools.escapeXml(subject)).append("</HN:subject>\n");
        }

        if (body != null) {
            fields.append(" <HN:textdescription>").append(StringTools.escapeXml(body)).append("</HN:textdescription>\n");
        }

        if (date != null) {
            fields.append(" <HN:date>").append(date).append("</HN:date>\n");
        }

        return MessageFormat.format(WEBDAV_MSG_PROPPATCH_UPDATE_NOTE,
                new Object[]{fields.toString()});
    }
}
