/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.security;

import java.security.Principal;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.core.Authentication;
import com.funambol.framework.core.Cred;
import com.funambol.framework.security.Officer;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.store.NotFoundException;
import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.tools.Base64;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.filter.WhereClause;

import com.funambol.server.config.Configuration;
import com.funambol.server.admin.UserManager;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.util.Def;
import com.funambol.exchange.admin.ExchangeConnectorConfig;
import com.funambol.exchange.items.common.manager.ItemManager;

/**
 * This is an implementation of the <i>Officier</i> interface.
 * It always authenticates and authorizes users and resource accesses.
 *
 * @version $Id: ExchangeOfficer.java,v 1.1 2006/08/08 07:46:31 gibi Exp $
 */
public class ExchangeOfficer implements Officer, java.io.Serializable {

    private static final String DEFAULT_USER_ROLE = "sync_user";

    // ---------------------------------------------------------- Protected data

    protected Logger log = Sync4jLogger.getLogger(Def.LOGGER_NAME);

    protected PersistentStore ps = null;

    protected UserManager  userManager = null;

    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for incorrect login/password?
     */
    private boolean loginFailed = false;

    public boolean isLoginFailed() {
        return loginFailed;
    }

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    /**
     * Which type of authetication does impose the server?
     */
    private String clientAuth = Cred.AUTH_TYPE_BASIC;

    public String getClientAuth() {
        return this.clientAuth;
    }

    public void setClientAuth(String clientAuth) {
        this.clientAuth = clientAuth;
    }

    /**
     * Which type of authetication use the server to send his credential?
     */
    private String serverAuth = Cred.AUTH_NONE;

    public String getServerAuth() {
        return this.serverAuth;
    }

    public void setServerAuth(String serverAuth) {
        this.serverAuth = serverAuth;
    }

    /**
     * Exchange server name
     */
    private String exchangeRoot = null;

    public String getExchangeRoot()
    {
        return exchangeRoot;
    }

    public void setExchangeRoot(String exchangeRoot)
    {
        this.exchangeRoot = exchangeRoot;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Authenticates a credential.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Cred credential) {

        if (log.isLoggable(Level.FINEST)) {
            log.finest(" ExchangeOfficer authenticate() Start");
        }

        Configuration config = Configuration.getConfiguration();
        ps = config.getStore();
        userManager = config.getUserManager();
        String type = credential.getType();

        if ((Cred.AUTH_TYPE_BASIC).equals(type)) {
            return authenticateBasicCredential(credential);
        }
        return false;
    }

    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource the name (or the identifier) of the resource to be authorized
     *
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /**
     * Un-authenticates a credential.
     * Do nothing. In the current implementation, the authentication is
     * discarde as soon as the LoginContext is garbage collected.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Cred credential) {
    }

    /**
     * @see com.funambol.framework.security.Officer
     *
     * @return boolean
     */
    public boolean isAccountExpired() {
        return loginExpired;
    }

    // --------------------------------------------------------- Private Methods

    /**
     *
     */
    private boolean authenticateBasicCredential(Cred credential)
    {

        String username;

        Authentication auth = credential.getAuthentication();
        String deviceId     = auth.getDeviceId();
        String credentials  = auth.getData();
        String userpwd      = new String(Base64.decode(auth.getData()));

        int p = userpwd.indexOf(':');

        if (p == -1)
        {
            username = userpwd;
        }
        else
        {
            username = (p > 0) ? userpwd.substring(0, p) : "";
        }

        if (log.isLoggable(Level.FINEST))
        {
            log.finest("Username: "    + username    );
            log.finest("Credentials: " + credentials );
        }

        String host ;
        int    port ;
        String exchangeHost ;

        try
        {
            host         = ExchangeConnectorConfig.getConfigInstance ().getSeccServer     ();
            port         = ExchangeConnectorConfig.getConfigInstance ().getSeccPort       ();
            exchangeHost = ExchangeConnectorConfig.getConfigInstance ().getExchangeServer ();
        }
        catch (Exception e)
        {
            log.severe("Error reading ExchangeConnectorConfig: " + e);
            log.throwing(getClass().getName(), "authenticateBasicCredential", e);
            return false;
        }

        //
        // Verify these credentials on Exchange
        //
        boolean isExcahngeUser = checkExchangeCredentials(host, port, 
                                                          exchangeHost, exchangeRoot, 
                                                          username, credentials);
        if (!isExcahngeUser) {
            return false;
        }

        try
        {
            if (existsUser(username)) {
                if (log.isLoggable(Level.INFO)) {
                    log.info("User with '" + username + "' exists.");
                }
            }
            else {
                insertUser(username, " ");
                if (log.isLoggable(Level.INFO)){
                    log.info("User '" + username + "' created.");
                }
            }
        } catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error inserting a new user: " + e.getMessage());
            } 
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, "authenticateBasicCredential", e);
            }
            return false;
        }

        try
        {
            if (!existsPrincipal(username, deviceId))
            {
                credential.getAuthentication().setPrincipalId(insertPrincipal(username, deviceId));
            }
        }
        catch(PersistentStoreException e)
        {
             if (log.isLoggable(Level.SEVERE))
            {
                log.severe("Error inserting a new principal: " + e.getMessage());
            }
            if (log.isLoggable(Level.INFO))
            {
                log.log(Level.INFO, "authenticateBasicCredential", e);
            }
            return false;
        }

        return true;
    }

    /**
     * Checks if the given login and password are valid
     * credentials on exchange server
     *
     *
     * @param host String
     * @param port int
     * @param exchangeHost String
     * @param exchangeServerName String
     * @param userName String
     * @param credentials String
     *
     *
     * @return boolean
     */
    private boolean checkExchangeCredentials(String host   ,
                                             int    port   ,
                                             String exchangeHost,
                                             String exchangeServerName ,
                                             String userName       ,
                                             String credentials) {
        int status;


        if (exchangeServerName == null || exchangeServerName.equals("")) {

            String message = "'Exchange server name' is null or empty. You have to set " +
                             "the Exchange server name in the file: " +
                             "Funambol/ds-server/config/funambol/server/security/ExchangeOfficer.xml";

            if (log.isLoggable(Level.FINEST)) {
                log.finest(message);
            }

            throw new IllegalStateException(message);
        }

        boolean check = false;

        try {

            ItemManager im = new ItemManager (host,
                                              port,
                                              exchangeHost);

            status = im.getExchangeAccessStatus(exchangeServerName ,
                                                userName           ,
                                                credentials        );

             if (status == 401 || 
                 status == 406 ||
                 status == 400 ) {
                check = false;
             } else {
                check = true;
             }

        } catch (DataAccessException dae) {
            log.severe("Error getting connection: " + dae);
            log.throwing(getClass().getName(), "check Exchange Credentials",
                         dae);
        }
        return check;
    }

    /**
         * Checks if there is an user with the given username
         * @param userName the username
         * @return true if the user exists, false otherwise
         */
        private boolean existsUser(String userName)
            throws PersistentStoreException
        {
            Sync4jUser[] users;
            WhereClause wc;
            String value[] = new String[]{userName};
            wc = new WhereClause("username", value, WhereClause.OPT_EQ, true);
            users = userManager.getUsers(wc);
            return users.length>0;
        }

    /**
     * Insert a new user with the given username and password
     * @param userName the username
     * @param password the password
     * @throws PersistentStoreException if an error occurs
     */
    private void insertUser(String userName, String password)
        throws PersistentStoreException
    {
        Sync4jUser user = new Sync4jUser();
        user.setUsername(userName);
        user.setPassword(password);
        user.setRoles(new String[] {DEFAULT_USER_ROLE});
        userManager.insertUser(user);
    }

    /**
     * Verify if there is a principal with the given userName and deviceId in store
     * @param userName the username
     * @param deviceId the device's id
     * @throws PersistentStoreException if an error occurs
     */
    private boolean existsPrincipal(String userName, String deviceId)
        throws PersistentStoreException
    {
        Principal principal;
        try
        {
            principal = Sync4jPrincipal.createPrincipal(userName, deviceId);
            ps.read(principal);
            return true;
        }
        catch(NotFoundException nfe)
        {
            if (log.isLoggable(Level.FINEST))
            {
                log.finest("Principal for " + userName + ":" + deviceId +" not found");
            }
            return false;
        }

     }

    /**
     * Insert a principal
     * @param userName the username
     * @param deviceId the device's id
     * @return the id of the principal
     * @throws PersistentStoreException if an error occurs
     */
    private long insertPrincipal(String userName, String deviceId)
        throws PersistentStoreException
    {
        Sync4jPrincipal principal = Sync4jPrincipal.createPrincipal(userName, deviceId);
        try
        {
            ps.store(principal);
        }
        catch (PersistentStoreException ex)
        {
            ex.printStackTrace();
            if (log.isLoggable(Level.SEVERE))
            {
                log.severe("Error creating new principal: " + ex);
            }
            if (log.isLoggable(Level.INFO))
            {
                log.log(Level.INFO, "authenticateBasicCredential",  ex);
            }
            throw ex;
        }
        return principal.getId();
    }



}
