/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.util;

import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.TimeZone;

import com.funambol.foundation.pdi.utils.TimeUtils;

import com.funambol.exchange.xml.XmlParseException;

/**
 * Utility class that groups dates manipulation functions.
 *
 * @version $Id: DateTools.java,v 1.1 2006/08/08 07:46:13 gibi Exp $
 */
public class DateTools {

    // ---------------------------------------------------------------- Contants
    private static final String DATE_FORMAT_WEBDAV =
            "yyyy-MM-dd'T'HH:mm:ss'.000Z'";

    private static final String DATE_FORMAT_PDI =
            "yyyyMMdd'T'HHmmss'Z'";

    private static final String ANNIVERSARY_FORMAT_PDI =
            "yyyy-MM-dd";

    private static final String ANNIVERSARY_FORMAT_WEBDAV =
            "yyyy-MM-dd'T'HH:mm:ss'.000Z'";

    private static final String DATETIME_FORMAT_CLIENT_UTC =
            "yyyyMMdd'T'HHmmss'Z'";

    private static final String DATETIME_FORMAT_CLIENT_NO_UTC =
            "yyyyMMdd'T'HHmmss";

    private static final String DATE_FORMAT_WEBDAV_FROM_EXCHANGE =
            "yyyy-MM-dd HH:mm:ss.SSS";

    protected static final int MAX_YEAR = 4000;


    /**
     * Make a webdav tag date from a date
     *
     * @param date     date
     * @param timeZone
     * @return webdav tag date
     * @throws XmlParseException
     */
    public static String dateToWebDavTag(String date, TimeZone timeZone)
            throws Exception {

        SimpleDateFormat formatter;
        String webDavDate;
        Date dt;

        if (date != null) {

            formatter = new SimpleDateFormat(DATE_FORMAT_PDI);
            formatter.setTimeZone(timeZone);
            dt = formatter.parse(date);

            webDavDate = dateToWebDavTag(dt);

        } else {
            webDavDate = "";
        }

        return webDavDate;
    }

    /**
     * Make a webdav tag date from a date
     *
     * @param date date
     * @return webdav tag date
     */
    public static String dateToWebDavTag(Date date) {

        SimpleDateFormat formatter;
        String webDavDate;

        if (date != null) {

            formatter = new SimpleDateFormat(DATE_FORMAT_WEBDAV);
            formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
            webDavDate = formatter.format(date);

            if (webDavDate.length() > 4) {
                String year = webDavDate.substring(0, 4);
                if (Integer.parseInt(year) > MAX_YEAR) {
                    webDavDate = "";
                }
            }
        } else {
            webDavDate = "";
        }

        return webDavDate;
    }


    /**
     * Make a webdav tag date from a date in day format.
     *
     * @param date date
     * @param tz   timezone
     * @return webdav tag date
     * @throws XmlParseException
     */
    public static String dateInDayFormatToWebDavTag(String date,
                                                    TimeZone tz,
                                                    boolean isAStartDate) throws Exception {

        if (isAStartDate) {
            //
            // The start date has time 00:00:00
            //
            date = TimeUtils.convertDateFromInDayFormat(date, "000000", false);
        } else {
            //
            // The end date has time 23:59:00
            //
            date = TimeUtils.convertDateFromInDayFormat(date, "240000", false);
        }

        date = TimeUtils.convertLocalDateToUTC(date, tz);

        return dateToWebDavTag(date, TimeZone.getDefault());
    }


    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     */
    public static Date pdiToDate(String date) throws XmlParseException {

        SimpleDateFormat formatter;
        Date dt;

        try {

            if (date != null) {

                formatter = new SimpleDateFormat(DATE_FORMAT_PDI);

                dt = formatter.parse(date);
            } else {
                dt = null;
            }
        } catch (ParseException e) {
            throw new XmlParseException(e.toString());
        }

        return dt;

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     */
    public static String webDavToPDIAnniversary(String date) throws Exception {

        SimpleDateFormat formatter;

        Date dt;
        String pdiDate;

        if (date != null && date.length() > 0) {

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_WEBDAV);

            dt = formatter.parse(date);

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_PDI);

            pdiDate = formatter.format(dt);

        } else {
            pdiDate = null;
        }

        return pdiDate;

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     */
    public static String PDIToWebDavAnniversary(String date) throws Exception {

        SimpleDateFormat formatter;

        Date dt;
        String pdiDate;

        if (date != null && date.length() > 0) {

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_PDI);

            dt = formatter.parse(date);

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_WEBDAV);

            pdiDate = formatter.format(dt);

        } else {
            pdiDate = null;
        }

        return pdiDate;

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     */
    public static Date webDavTagToDate(String date)
            throws XmlParseException {

        SimpleDateFormat formatter;
        String value;
        Date timestamp;

        try {

            if (date != null && date.length()>23) {
                value = date.substring(0, 10) +
                        " " +
                        date.substring(11, 23);

                formatter = new SimpleDateFormat(DATE_FORMAT_WEBDAV_FROM_EXCHANGE);
                formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
                timestamp = formatter.parse(value);
            } else {
                timestamp = null;
            }
        } catch (ParseException e) {
            throw new XmlParseException(e.toString());
        }

        return timestamp;

    }

    /**
     * Make a date from a webdav tag date to client date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     */
    public static String toClientDate(Date date)
            throws XmlParseException {

        SimpleDateFormat formatter;
        try {

            if (date != null) {

                formatter = new SimpleDateFormat(DATETIME_FORMAT_CLIENT_UTC);
                formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
                return formatter.format(date);

            } else {
                return null;
            }
        } catch (Exception e) {
            throw new XmlParseException(e.toString());
        }

    }

    /**
     * Make a date from a client date to webdav date
     *
     * @param date client date
     * @return a Date
     * @throws XmlParseException
     */
    public static Date clientDateToDate(String date)
            throws XmlParseException {

        SimpleDateFormat formatter;
        Date timestamp;

        String format;

        try {

            if (date != null) {

                if (date.indexOf("Z") != -1) {
                    format = DATETIME_FORMAT_CLIENT_UTC;
                } else {
                    //
                    // for nokia phone utc bug
                    //
                    format = DATETIME_FORMAT_CLIENT_NO_UTC;
                }

                formatter = new SimpleDateFormat(format);
                formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
                timestamp = formatter.parse(date);

                return timestamp;

            } else {
                return null;
            }
        } catch (ParseException e) {
            throw new XmlParseException(e.toString());
        }
    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date     webdav tag content
     * @param timeZone
     * @return a Date
     * @throws XmlParseException
     */
    public static String webDavTagToPDIDate(String date, TimeZone timeZone)
            throws XmlParseException {

        SimpleDateFormat formatter;

        if (timeZone==null){
            timeZone = TimeZone.getTimeZone("GMT");
        }

        Date dt;
        String pdiDate;

        if (date != null && date.length() > 0) {

            dt = webDavTagToDate(date);

            formatter = new SimpleDateFormat(DATE_FORMAT_PDI);
            formatter.setTimeZone(timeZone);
            pdiDate = formatter.format(dt);

        } else {
            pdiDate = null;
        }

        return pdiDate;
    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     */
    public static String dateToPDIDate(Date date) {

        SimpleDateFormat formatter;

        String pdiDate;

        if (date != null) {

            formatter = new SimpleDateFormat(DATE_FORMAT_PDI);

            pdiDate = formatter.format(date);

        } else {
            pdiDate = null;
        }

        return pdiDate;
    }


    /**
     * <p> Convert a date in WebDav format ("yyyy-MM-dd'T'KK:mm:ss.fff'Z'")
     * in day format ("yyyy-MM-dd") using the input timezone. </p>
     * @param date
     * @param timezone
     * @param isAStartDate flag that tells if the input String represents a start date
     * @return the date in day format
     * @throws XmlParseException
     */
    public static String convertWebDavDateInDayFormat(String date,
                                                      TimeZone timezone,
                                                      boolean isAStartDate)
    throws XmlParseException
    {
        SimpleDateFormat formatter;

        Date dt;
        String dayDate, value;

        if (date != null && date.length() > 0)
        {
            value = date;

            formatter = new SimpleDateFormat(DATE_FORMAT_WEBDAV);
            formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
            try
            {
                dt = formatter.parse(value);
            }
            catch(ParseException e)
            {
                throw  new XmlParseException(e.getMessage());
            }

            formatter.applyPattern(ANNIVERSARY_FORMAT_PDI);

            if (timezone==null){
                timezone = TimeZone.getTimeZone("GMT");
            }

            formatter.setTimeZone(timezone);

            if (!isAStartDate)
            {
                dt.setTime(dt.getTime() - 24*60*60*1000);
            }


            dayDate = formatter.format(dt);
        }
        else
        {
            dayDate = null;
        }

        return dayDate;
    }

    /**
     * <p>Converts a date in day format("yyyy-MM-dd") to WebDAV format ("yyyy-MM-dd'T'KK:mm:ss'.000Z'").</p>
     * <p>If the timezone parameter is <i>null</i> the timezone of the output date is considered GMT.</p>
     * <p>If the isAStartDate is true then the time in GMT of the output date will be 00:00:00 else 24:00:00.</p>
     * @param date
     * @param timezone
     * @param isAStartDate
     * @return the date in WebDav format
     * @throws XmlParseException
     */
    public static String convertDayInWebDavDateFormat(String date,
                                                      TimeZone timezone,
                                                      boolean isAStartDate)
            throws XmlParseException {
        SimpleDateFormat formatter;

        Date dt;
        String wedDavDate, value;

        if (date != null && date.length() > 0) {
            if (isAStartDate){
                value = date.substring(0, 10)+"T00:00:00.000Z";
            }
            else{
                value = date.substring(0, 10)+"T24:00:00.000Z";
            }

            formatter = new SimpleDateFormat(DATE_FORMAT_WEBDAV);
            if (timezone == null) {
                timezone = TimeZone.getTimeZone("GMT");
            }
            formatter.setTimeZone(timezone);
            try {
                dt = formatter.parse(value);
            }
            catch (ParseException e) {
                throw new XmlParseException(e.getMessage());
            }
            formatter.setTimeZone(TimeZone.getTimeZone("GMT"));

            wedDavDate = formatter.format(dt);

        } else {
            wedDavDate = null;
        }

        return wedDavDate;
    }
}
